// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0
package jobqueue

import (
	"github.com/stretchr/testify/assert"
	"gorm.io/driver/sqlite"
	"gorm.io/gorm"
	"testing"
	"time"
)

func TestWriteToDB4(t *testing.T) {

	db, err := gorm.Open(sqlite.Open("file::memory:?cache=shared"), &gorm.Config{})
	if err != nil {
		t.Fatalf("a error occurred while opening the database: %v", err)
	}

	manager := NewManager()
	manager.SetDB(db)
	err = manager.AddWorker(NewLocalWorker(1))
	assert.Nil(t, err)

	err = manager.Start()
	assert.Nil(t, err)

	runner := &CounterRunnable{}
	job := NewJob[CounterResult]("job3", runner)

	job.stats = JobStats{
		JobID:        job.GetID(),
		RunCount:     20,
		SuccessCount: 30,
		ErrorCount:   40,
	}

	scheduler := &InstantScheduler{}
	err = manager.ScheduleJob(job, scheduler)
	assert.Nil(t, err)

	time.Sleep(500 * time.Millisecond)

	// check is stats are the values above
	var tmpJob JobPersistence

	err = db.Preload("Stats").First(&tmpJob, "id = ?", "job3").Error
	assert.Nil(t, err)

	// Validate the fields
	assert.Equal(t, JobID("job3"), tmpJob.ID)
	assert.Equal(t, 21, tmpJob.Stats.RunCount)
	assert.Equal(t, 31, tmpJob.Stats.SuccessCount)
	assert.Equal(t, 40, tmpJob.Stats.ErrorCount)

	// reset stats
	err = manager.ResetJobStats(job.GetID())
	assert.Nil(t, err)

	time.Sleep(500 * time.Millisecond)

	var tmpJob2 JobPersistence
	// check is stats are the values above
	err = db.First(&tmpJob2, "id = ?", "job3").Error
	err = db.Preload("Stats").First(&tmpJob2, "id = ?", "job3").Error
	assert.Nil(t, err)

	// Validate the fields
	assert.Equal(t, JobID("job3"), tmpJob2.ID)
	assert.Equal(t, 0, tmpJob2.Stats.RunCount)
	assert.Equal(t, 0, tmpJob2.Stats.SuccessCount)
	assert.Equal(t, 0, tmpJob2.Stats.ErrorCount)

	err = manager.DeleteJob(job.GetID())
	assert.Nil(t, err)

}
