// Copyright 2023 schukai GmbH
// SPDX-License-Identifier: AGPL-3.0

package jobqueue

import (
	"sync"
)

type JobSyncer struct {
	jobQueue        []GenericJob
	queueLock       sync.Mutex
	notifyChannel   chan struct{}
	stopChan        chan struct{}
	jobSaveProgress sync.WaitGroup
	status          Status
	manager         *Manager
	mu              sync.Mutex
	migrateFlag     bool
}

type Status int

const (
	JobSyncerStatusStopped Status = iota
	JobSyncerStatusRunning
)

func CreateAndStartJobSyncer[P *Promise[*JobSyncer]](manager *Manager) *Promise[*JobSyncer] {

	s := NewJobSyncer(manager)

	s.mu.Lock()
	defer s.mu.Unlock()

	return NewPromise[*JobSyncer](func(resolve func(*JobSyncer), reject func(error)) {

		if s.manager == nil || s.manager.database == nil {
			reject(ErrNoDatabaseConnection)
			return
		}

		if s.status == JobSyncerStatusRunning {
			resolve(s)
			return
		}

		db := s.manager.database
		if !s.migrateFlag {
			err := db.AutoMigrate(&JobPersistence{}, &JobLog{}, &JobStats{})
			if err != nil {
				reject(err)
				return
			}
			s.migrateFlag = true
		}

		err := s.Start()
		if err != nil {
			reject(err)
			return
		}
		resolve(s)
	})
}

func NewJobSyncer(manager *Manager) *JobSyncer {
	return &JobSyncer{
		jobQueue: make([]GenericJob, 0),
		manager:  manager,
	}
}

func (js *JobSyncer) Start() error {
	js.mu.Lock()
	defer js.mu.Unlock()

	if js.status == JobSyncerStatusRunning {
		return ErrJobSyncerAlreadyRunning
	}

	js.notifyChannel = make(chan struct{}, 1) // Buffer to avoid blocking
	js.stopChan = make(chan struct{})
	js.status = JobSyncerStatusRunning

	go js.runWorker()
	return nil
}

func (js *JobSyncer) runWorker() {
	for {
		select {
		case <-js.notifyChannel:
			js.processJobs()
		case <-js.stopChan:
			js.cleanup()
			return
		}
	}
}

func (js *JobSyncer) processJobs() {
	for {
		js.queueLock.Lock()
		if len(js.jobQueue) == 0 {
			js.queueLock.Unlock()
			return
		}
		job := js.jobQueue[0]
		js.jobQueue = js.jobQueue[1:]
		js.queueLock.Unlock()

		js.jobSaveProgress.Add(1)
		js.processJob(job)
		js.jobSaveProgress.Done()
	}
}

func (js *JobSyncer) AddJob(job GenericJob) {
	js.queueLock.Lock()

	// check if job is already in queue
	for _, j := range js.jobQueue {
		if j.GetID() == job.GetID() {
			js.queueLock.Unlock()
			return
		}
	}

	js.jobQueue = append(js.jobQueue, job)
	js.queueLock.Unlock()

	js.jobSaveProgress.Add(1)

	// Non-blocking notify
	select {
	case js.notifyChannel <- struct{}{}:
	default:
	}
}

func (js *JobSyncer) processJob(job GenericJob) {
	defer js.jobSaveProgress.Done()
	err := createOrUpdateJob(job, js.manager.database)
	Error("Error while creating or updating job", "error", err)
}

func (js *JobSyncer) Stop() error {
	js.mu.Lock()
	if js.status != JobSyncerStatusRunning {
		js.mu.Unlock()
		return ErrJobSyncerNotRunning
	}
	js.status = JobSyncerStatusStopped
	js.mu.Unlock()

	close(js.stopChan)
	js.jobSaveProgress.Wait()

	return nil
}

func (js *JobSyncer) cleanup() {
	js.mu.Lock()
	defer js.mu.Unlock()

	js.jobSaveProgress.Wait()
	js.status = JobSyncerStatusStopped
}
